/*
 ThermostatQA - test framework for Thermostat Monitoring Tool

 Copyright 2015 Red Hat, Inc.

 This file is part of ThermostatQA

 ThermostatQA is distributed under the GNU General Public License,
 version 2 or any later version (with a special exception described
 below, commonly known as the "Classpath Exception").

 A copy of GNU General Public License (GPL) is included in this
 distribution, in the file COPYING.

 Linking ThermostatQA code with other modules is making a combined work
 based on ThermostatQA.  Thus, the terms and conditions of the GPL
 cover the whole combination.

 As a special exception, the copyright holders of ThermostatQA give you
 permission to link this code with independent modules to produce an
 executable, regardless of the license terms of these independent
 modules, and to copy and distribute the resulting executable under
 terms of your choice, provided that you also meet, for each linked
 independent module, the terms and conditions of the license of that
 module.  An independent module is a module which is not derived from
 or based on ThermostatQA code.  If you modify ThermostatQA, you may
 extend this exception to your version of the software, but you are
 not obligated to do so.  If you do not wish to do so, delete this
 exception statement from your version.
 */
package org.thermostat.qa2.reporter;

import org.thermostat.qa2.reporter.result.TestMethodResult;
import org.thermostat.qa2.reporter.result.TestClassResult;
import org.thermostat.qa2.reporter.result.TestRunResult;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.thermostat.qa2.framework.TestRunner;
import org.thermostat.qa2.framework.TestResult;
import org.thermostat.qa2.framework.utils.FileUtilities;

/**
 *
 * @author Zdeněk Žamberský
 */
public class Generator {

    public static final String resultsUrlPattern = "${RESULTS_URL}";
    public static final String datePattern = "${DATE}";
    public static final String testCountPattern = "${TEST_COUNT}";

    public static final String passedPattern = "${PASSED}";
    public static final String failedPattern = "${FAILED}";
    public static final String errorPattern = "${ERROR}";
    public static final String ignoredPattern = "${IGNORED}";

    public static final String osNamePattern = "${OS_NAME}";
    public static final String osVersionPattern = "${OS_VER}";
    public static final String osArchPattern = "${OS_ARCH}";

    public static final String javaVersionPattern = "${JAVA_VERSION}";
    public static final String vmNamePattern = "${VM_NAME}";
    public static final String vmVersionPattern = "${VM_VERSION}";

    public static final String sumaryPattern = "${SUMMARY}";
    public static final String resultsPattern = "${RESULTS}";
    public static final String tableDataPattern = "${TABLE_DATA}";

    public static final String graphDataNonePattern = "${GRAPH_DATA_NONE}";
    public static final String graphDataPassedPattern = "${GRAPH_DATA_PASSED}";
    public static final String graphDataFailedPattern = "${GRAPH_DATA_FAILED}";
    public static final String graphDataErrorPattern = "${GRAPH_DATA_ERROR}";
    public static final String graphDataIgnoredPattern = "${GRAPH_DATA_IGNORED}";

    public static final String indexPageTemplate = "templates/index.html";
    public static final String logPageTemplate = "templates/log.html";
    public static final String historyPageTemplate = "templates/hist.html";
    public static final String graphPageTemplate = "templates/graph.html";

    ////////////////////
    //// INDEX PAGE ////
    ////////////////////
    public static void generateIndexPage(List<String> output, TestRunResult run) throws IOException {
        Map<String, List<String>> replacements = new HashMap();
        addReplacementToMap(replacements, resultsUrlPattern, "log_" + run.date + ".xhtml");
        addReplacementToMap(replacements, datePattern, run.date);
        addReplacementToMap(replacements, passedPattern, Integer.toString(run.passedCount));
        addReplacementToMap(replacements, failedPattern, Integer.toString(run.failedCount));
        addReplacementToMap(replacements, errorPattern, Integer.toString(run.errorCount));
        addReplacementToMap(replacements, ignoredPattern, Integer.toString(run.ignoredCount));
        addReplacementToMap(replacements, osNamePattern, System.getProperty("os.name"));
        addReplacementToMap(replacements, osVersionPattern, System.getProperty("os.version"));
        addReplacementToMap(replacements, osArchPattern, System.getProperty("os.arch"));
        addReplacementToMap(replacements, javaVersionPattern, System.getProperty("java.version"));
        addReplacementToMap(replacements, vmNamePattern, System.getProperty("java.vm.name"));
        addReplacementToMap(replacements, vmVersionPattern, System.getProperty("java.vm.version"));
        List<String> template = FileUtilities.getLineListFromFile(indexPageTemplate);
        replacePatterns(template, output, replacements);
    }

    //////////////////
    //// LOG PAGE ////
    //////////////////
    public static void generateLogPage(List<String> output, TestRunResult run) throws IOException {
        List<String> dateText = new ArrayList();
        dateText.add(run.date);
        List<String> summaryText = new ArrayList();
        generateLogSummary(summaryText, run);
        List<String> resultsText = new ArrayList();
        generateLogTable(resultsText, run);
        Map<String, List<String>> replacements = new HashMap();
        replacements.put(datePattern, dateText);
        replacements.put(sumaryPattern, summaryText);
        replacements.put(resultsPattern, resultsText);
        List<String> template = FileUtilities.getLineListFromFile(logPageTemplate);
        replacePatterns(template, output, replacements);
    }

    public static void generateLogSummary(List<String> output, TestRunResult run) {
        for (TestClassResult test : run.results) {
            output.add("<tr><td><a href='#" + test.name + "'>" + test.name + "</a></td></tr>");
        }
    }

    public static void generateLogTable(List<String> output, TestRunResult run) {
        for (TestClassResult test : run.results) {
            String testName = test.name;
            output.add("<tr id='" + testName + "' ><td class='table-header' colspan='2'>" + testName + "</td></tr>");
            for (TestMethodResult method : test.methods) {
                TestResult result = method.result;
                String resultString = result.toString();
                String resultLine = "<tr id='" + testName + "." + method.name + "' >";
                resultLine += result == TestResult.PASSED ? "<td>" : "<td class='" + resultString.toLowerCase() + "-header'" + ">";
                resultLine += "<span class='test-name-prefix'>" + testName + ".</span><span class='test-name-postfix'>" + method.name + "</span>";
                if (method.reason != null) {
                    resultLine += "<span class='test-name-postfix'>:&nbsp;&nbsp;</span><small class='reason'>" + method.reason + "</small>";
                }
                resultLine += "</td><td class='" + resultString.toLowerCase() + "-text'>" + resultString + "</td></tr>";
                output.add(resultLine);
                List<String> stackTrace = method.stackTrace;
                if (stackTrace != null) {
                    output.add("<tr><td>");
                    output.add("<input id=\"log-toggle-" + testName + "-" + method.name + "\" type=\"checkbox\" class=\"log-toggle\" />");
                    output.add("<label for=\"log-toggle-" + testName + "-" + method.name + "\"></label><br />");
                    output.add("<div>");
                    output.add("<h5>stack trace:</h5>");
                    output.add("<pre class='stack-trace'><![CDATA[" + (stackTrace.size() > 0 ? stackTrace.get(0) : ""));
                    for (int i = 1; i < stackTrace.size(); ++i) {
                        output.add(stackTrace.get(i));
                    }
                    output.add("]]></pre>");
                    output.add("<h5>log:</h5>");
                    List<String> log = method.getLog();
                    output.add("<pre class='stack-trace'><![CDATA[" + (log.size() > 0 ? log.get(0) : ""));
                    for (int i = 1; i < log.size(); ++i) {
                        output.add(log.get(i));
                    }
                    output.add("]]></pre>");
                    output.add("</div>");
                    output.add("</td></tr>");
                }
            }
            output.add("<tr><td class='error-text' colspan='2'>SUMMARY: " + testName + "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;total: " + test.methods.size() + "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;passed: " + test.passedCount + "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;failed: " + test.failedCount + "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;error: " + test.errorCount + "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;not applicable: " + test.ignoredCount + "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;duration: " + test.duration + "</td></tr>");
            output.add("<tr><td colspan='2'>&nbsp;</td></tr>");
        }
    }

    //////////////////////
    //// HISTORY PAGE ////
    //////////////////////
    public static void generateHistoryPage(List<String> output, List<TestRunResult> runs, String resultsType, int count, boolean onlyFail) throws IOException {
        List<String> resultTypeText = new ArrayList();
        resultTypeText.add(resultsType);
        List<String> historyTableText = new ArrayList();
        generateHistoryTable(historyTableText, runs, count, onlyFail);
        Map<String, List<String>> replacements = new HashMap();
        replacements.put(resultsPattern, resultTypeText);
        replacements.put(tableDataPattern, historyTableText);
        List<String> template = FileUtilities.getLineListFromFile(historyPageTemplate);
        replacePatterns(template, output, replacements);
    }

    public static void generateHistoryTable(List<String> output, List<TestRunResult> runs, int count, boolean onlyFail) {
        int size = runs.size();
        int start = Math.max(0, size - count);
        TestRunResult lastRun = runs.get(size - 1);
        output.add("<tr><td class='group-id'>Test name/Date:</td>");
        for (int i = start; i < size; ++i) {
            String name = runs.get(i).date;
            int index = name.indexOf("-");
            String year = name.substring(0, index);
            String date = name.substring(index + 1, name.length());
            output.add("<td>" + year + "<br />" + date + "</td>");
        }
        output.add("</tr>");
        List<TestClassResult> lastResults = lastRun.results;
        List<TestClassResult> testHistory = new ArrayList();
        for (TestClassResult result : lastResults) {
            String name = result.name;
            for (int i = start; i < size; ++i) {
                Map<String, TestClassResult> testMap = runs.get(i).resultsMap;
                testHistory.add(testMap.get(name));
            }
            generateHistoryTableRow(output, runs, testHistory, count, onlyFail);
            testHistory.clear();
        }
    }

    public static void generateHistoryTableRow(List<String> output, List<TestRunResult> runs, List<TestClassResult> testHistory, int count, boolean onlyFail) {
        int size = testHistory.size();
        TestRunResult lastRun = runs.get(size - 1);
        TestClassResult current = testHistory.get(size - 1);
        List<TestMethodResult> allMethods = new ArrayList();
        for (TestMethodResult method : current.methods) {
            String name = method.getName();
            boolean fail = false;
            for (TestClassResult previous : testHistory) {
                if (previous != null) {
                    TestMethodResult previousMethod = previous.getMethod(name);
                    if (previousMethod != null && previousMethod.getResult() != TestResult.PASSED) {
                        fail = true;
                    }
                    allMethods.add(previousMethod);
                } else {
                    allMethods.add(null);
                }
            }
            if (!onlyFail || fail) {
                String lastTestFile = "log_" + lastRun.date + ".xhtml";
                String methodref = "#" + current.name + "." + method.name;
                output.add("<tr><td>" + current.name + ".<a href='" + lastTestFile + methodref + "'>" + method.getName() + "</a></td>");
                int i = 0;
                for (TestMethodResult printedMethod : allMethods) {
                    if (printedMethod != null) {
                        TestResult result = printedMethod.getResult();
                        String resultString = result.toString().toLowerCase();
                        String printedResultString = result == TestResult.PASSED ? "ok" : result == TestResult.FAILED ? "fail" : resultString;
                        output.add("<td class='" + resultString + "-header'><a href='" + "log_" + runs.get(i).date + ".xhtml" + methodref + "'>" + printedResultString + "</a></td>");
                    } else {
                        output.add("<td> </td>");
                    }
                    ++i;
                }
                output.add("</tr>");
            }
            allMethods.clear();
        }
    }

    ////////////////////
    //// GRAPH PAGE ////
    ////////////////////
    public static void generateGraphPage(List<String> output, List<TestRunResult> runs, int count) throws IOException {
        List<String> testCountText = new ArrayList();
        testCountText.add(Integer.toString(count));
        List<String> passedDataText = new ArrayList();
        generateGraphData(passedDataText, runs, TestResult.PASSED, count);
        List<String> failedDataText = new ArrayList();
        generateGraphData(failedDataText, runs, TestResult.FAILED, count);
        List<String> errorDataText = new ArrayList();
        generateGraphData(errorDataText, runs, TestResult.ERROR, count);
        List<String> ignorredDataText = new ArrayList();
        generateGraphData(ignorredDataText, runs, TestResult.IGNORED, count);
        List<String> noneDataText = new ArrayList();
        generateGraphData(noneDataText, runs, null, count);
        List<String> tableText = new ArrayList();
        generateGraphTable(tableText, runs, count);
        Map<String, List<String>> replacements = new HashMap();
        replacements.put(testCountPattern, testCountText);
        replacements.put(graphDataPassedPattern, passedDataText);
        replacements.put(graphDataFailedPattern, failedDataText);
        replacements.put(graphDataErrorPattern, errorDataText);
        replacements.put(graphDataIgnoredPattern, ignorredDataText);
        replacements.put(graphDataNonePattern, noneDataText);
        replacements.put(tableDataPattern, tableText);
        List<String> template = FileUtilities.getLineListFromFile(graphPageTemplate);
        replacePatterns(template, output, replacements);
    }

    public static void generateGraphTable(List<String> output, List<TestRunResult> results, int count) {
        int size = results.size();
        int start = Math.max(0, size - count);
        for (int i = start; i < results.size(); ++i) {
            TestRunResult result = results.get(i);
            output.add("<tr><td>" + i + "</td><td><a href='log_" + result.date + ".xhtml'>&nbsp;" + result.date + "&nbsp;</a></td>");
            output.add("<td style='color:#006000;text-align:right;'>" + result.passedCount + "</td>");
            output.add("<td style='color:#800000;text-align:right;'>" + result.failedCount + "</td>");
            output.add("<td style='color:#000080;text-align:right;'>" + result.errorCount + "</td>");
            output.add("<td style='color:#606060;text-align:right;'>" + result.ignoredCount + "</td>");
            output.add("</tr>");
        }
    }

    public static void generateGraphData(List<String> output, List<TestRunResult> results, TestResult result, int count) {
        int size = results.size();
        int start = Math.max(0, size - count);
        for (int i = start; i < results.size(); ++i) {
            TestRunResult run = results.get(i);
            int targetResultCount = result == TestResult.PASSED ? run.passedCount : result == TestResult.FAILED ? run.failedCount : result == TestResult.ERROR ? run.errorCount : result == TestResult.IGNORED ? run.ignoredCount : 0;
            output.add("[" + (i + 1) + ", " + targetResultCount + "],");
        }
    }

    ////////////////
    //// COMMON ////
    ////////////////
    public static void replacePatterns(List<String> template, List<String> output, Map<String, List<String>> patterns) {
        int patternsCount = patterns.size();
        Map.Entry<String, List<String>>[] entries = new Map.Entry[patternsCount];
        patterns.entrySet().toArray(entries);
        for (String line : template) {
            String restOfLine = line;
            String toPrint = null;
            Map.Entry<String, List<String>> entryFound;
            do {
                entryFound = null;
                int index = -1;
                for (int i = 0; i < patternsCount; ++i) {
                    Map.Entry<String, List<String>> entry = entries[i];
                    String pattern = entry.getKey();
                    int foundIndex = restOfLine.indexOf(pattern);
                    if (foundIndex >= 0 && (entryFound == null || foundIndex < index)) {
                        index = foundIndex;
                        entryFound = entry;
                    }
                }
                if (entryFound != null) {
                    String pattern = entryFound.getKey();
                    List<String> replacement = entryFound.getValue();
                    int replacementSize = replacement.size();
                    toPrint = line.substring(0, index);
                    for (int j = 0; j < replacementSize - 1; ++j) {
                        String replacementLine = replacement.get(j);
                        output.add(toPrint == null ? replacementLine : toPrint + replacementLine);
                        toPrint = null;
                    }
                    String lastReplacementLine = replacementSize > 0 ? replacement.get(replacementSize - 1) : "";
                    toPrint = toPrint == null ? lastReplacementLine : toPrint + lastReplacementLine;
                    restOfLine = line.substring(index + pattern.length());
                }
            } while (entryFound != null);
            output.add(toPrint == null ? restOfLine : toPrint + restOfLine);
        }
    }

    public static void addReplacementToMap(Map<String, List<String>> map, String pattern, String replacement) {
        List<String> replacementText = new ArrayList();
        replacementText.add(replacement);
        map.put(pattern, replacementText);
    }

}
